@@ -99,7 +99,7 @@ static int hnbgw_tx_hnb_register_acc(struct hnb_context *ctx)
 
 	/* Single required response IE: RNC-ID */
 	HNBAP_HNBRegisterAcceptIEs_t accept = {
-		.rnc_id = ctx->gw->config.rnc_id
+		.rnc_id = g_hnbgw->config.rnc_id
 	};
 
 	/* encode the Information Elements */
@@ -119,7 +119,7 @@ static int hnbgw_tx_hnb_register_acc(struct hnb_context *ctx)
 
 	ASN_STRUCT_FREE_CONTENTS_ONLY(asn_DEF_HNBAP_HNBRegisterAccept, &accept_out);
 
-	LOGHNB(ctx, DHNBAP, LOGL_NOTICE, "Accepting HNB-REGISTER-REQ from %s\n", ctx->identity_info);
+	LOGHNB(ctx, DHNBAP, LOGL_NOTICE, "Accepting HNB-REGISTER-REQ\n");
 
 	return hnbgw_hnbap_tx(ctx, msg);
 }
@@ -146,6 +146,7 @@ static int hnbgw_tx_ue_register_acc(struct ue_context *ue)
 
 	memset(&accept_out, 0, sizeof(accept_out));
 	rc = hnbap_encode_ueregisteraccepties(&accept_out, &accept);
+	ASN_STRUCT_FREE_CONTENTS_ONLY(asn_DEF_OCTET_STRING, &accept.uE_Identity.choice.iMSI);
 	if (rc < 0) {
 		return rc;
 	}
@@ -155,7 +156,6 @@ static int hnbgw_tx_ue_register_acc(struct ue_context *ue)
 						&asn_DEF_HNBAP_UERegisterAccept,
 						&accept_out);
 
-	ASN_STRUCT_FREE_CONTENTS_ONLY(asn_DEF_OCTET_STRING, &accept.uE_Identity.choice.iMSI);
 	ASN_STRUCT_FREE_CONTENTS_ONLY(asn_DEF_HNBAP_UERegisterAccept, &accept_out);
 
 	return hnbgw_hnbap_tx(ue->hnb, msg);
@@ -323,14 +323,14 @@ static int hnbgw_tx_ue_register_acc_tmsi(struct hnb_context *hnb, HNBAP_UE_Ident
 		break;
 
 	default:
-		LOGHNB(hnb, DHNBAP, LOGL_ERROR, "Unsupportedccept UE ID (present=%d)\n", ue_id->present);
+		LOGHNB(hnb, DHNBAP, LOGL_ERROR, "Unsupported HNBAP UE Identity type (present=%d)\n", ue_id->present);
 		return -1;
 	}
 
 	tmsi = ntohl(tmsi);
 	LOGHNB(hnb, DHNBAP, LOGL_DEBUG, "HNBAP register with TMSI %x\n", tmsi);
 
-	ue = ue_context_by_tmsi(hnb->gw, tmsi);
+	ue = ue_context_by_tmsi(tmsi);
 	if (!ue)
 		ue = ue_allocated = ue_context_alloc(hnb, NULL, tmsi);
 
@@ -338,19 +338,8 @@ static int hnbgw_tx_ue_register_acc_tmsi(struct hnb_context *hnb, HNBAP_UE_Ident
 
 	memset(&accept_out, 0, sizeof(accept_out));
 	rc = hnbap_encode_ueregisteraccepties(&accept_out, &accept);
-	if (rc < 0) {
-		/* If we allocated the UE context but the UE REGISTER fails, get rid of it again: there will likely
-		 * never be a UE DE-REGISTER for this UE from the HNB, and the ue_context would linger forever. */
-		if (ue_allocated)
-			ue_context_free(ue_allocated);
-		return rc;
-	}
-
-	msg = hnbap_generate_successful_outcome(HNBAP_ProcedureCode_id_UERegister,
-						HNBAP_Criticality_reject,
-						&asn_DEF_HNBAP_UERegisterAccept,
-						&accept_out);
 
+	/* free 'accept', now encoded in 'accept_out' */
 	switch (ue_id->present) {
 	case HNBAP_UE_Identity_PR_tMSILAI:
 		ASN_STRUCT_FREE_CONTENTS_ONLY(asn_DEF_BIT_STRING,
@@ -377,9 +366,23 @@ static int hnbgw_tx_ue_register_acc_tmsi(struct hnb_context *hnb, HNBAP_UE_Ident
 		break;
 	}
 
-	ASN_STRUCT_FREE_CONTENTS_ONLY(asn_DEF_HNBAP_UERegisterAccept, &accept_out);
+	if (rc < 0) {
+		/* Encoding failed. Nothing in 'accept_out'. */
+		/* If we allocated the UE context but the UE REGISTER fails, get rid of it again: there will likely
+		 * never be a UE DE-REGISTER for this UE from the HNB, and the ue_context would linger forever. */
+		if (ue_allocated)
+			ue_context_free(ue_allocated);
+		return rc;
+	}
 
-	return hnbgw_hnbap_tx(hnb, msg);
+	/* Encoding successful, transmit, then free 'accept_out'. */
+	msg = hnbap_generate_successful_outcome(HNBAP_ProcedureCode_id_UERegister,
+						HNBAP_Criticality_reject,
+						&asn_DEF_HNBAP_UERegisterAccept,
+						&accept_out);
+	rc = hnbgw_hnbap_tx(hnb, msg);
+	ASN_STRUCT_FREE_CONTENTS_ONLY(asn_DEF_HNBAP_UERegisterAccept, &accept_out);
+	return rc;
 }
 
 static int hnbgw_rx_hnb_deregister(struct hnb_context *ctx, ANY_t *in)
@@ -406,14 +409,10 @@ static int hnbgw_rx_hnb_register_req(struct hnb_context *ctx, ANY_t *in)
 	int rc;
 	struct osmo_plmn_id plmn;
 	struct osmo_fd *ofd = osmo_stream_srv_get_ofd(ctx->conn);
-	char name[OSMO_SOCK_NAME_MAXLEN];
-
-	osmo_sock_get_name_buf(name, sizeof(name), ofd->fd);
 
 	rc = hnbap_decode_hnbregisterrequesties(&ies, in);
 	if (rc < 0) {
-		LOGHNB(ctx, DHNBAP, LOGL_ERROR, "Failure to decode HNB-REGISTER-REQ %s from %s: rc=%d\n",
-		       ctx->identity_info, name, rc);
+		LOGHNB(ctx, DHNBAP, LOGL_ERROR, "Failure to decode HNB-REGISTER-REQ: rc=%d\n", rc);
 		return rc;
 	}
 
@@ -428,7 +427,7 @@ static int hnbgw_rx_hnb_register_req(struct hnb_context *ctx, ANY_t *in)
 	ctx->id.mcc = plmn.mcc;
 	ctx->id.mnc = plmn.mnc;
 
-	llist_for_each_entry_safe(hnb, tmp, &ctx->gw->hnb_list, list) {
+	llist_for_each_entry_safe(hnb, tmp, &g_hnbgw->hnb_list, list) {
 		if (hnb->hnb_registered && ctx != hnb && memcmp(&ctx->id, &hnb->id, sizeof(ctx->id)) == 0) {
 			/* If it's coming from the same remote IP addr+port, then it must be our internal
 			 * fault (bug), and we release the old context to keep going... */
@@ -457,17 +456,21 @@ static int hnbgw_rx_hnb_register_req(struct hnb_context *ctx, ANY_t *in)
 
 			/* If new conn registering same HNB is from anoter remote addr+port, let's reject it to avoid
 			 * misconfigurations or someone trying to impersonate an already working HNB: */
-			LOGHNB(ctx, DHNBAP, LOGL_ERROR, "rejecting HNB-REGISTER-REQ with duplicate cell identity "
-				"MCC=%u,MNC=%u,LAC=%u,RAC=%u,SAC=%u,CID=%u from %s\n",
-				ctx->id.mcc, ctx->id.mnc, ctx->id.lac, ctx->id.rac, ctx->id.sac, ctx->id.cid, name);
+			LOGHNB(ctx, DHNBAP, LOGL_ERROR, "rejecting HNB-REGISTER-REQ with duplicate cell identity %s\n",
+			       umts_cell_id_name(&ctx->id));
 			hnbap_free_hnbregisterrequesties(&ies);
 			return hnbgw_tx_hnb_register_rej(ctx);
 		}
 	}
 
-	LOGHNB(ctx, DHNBAP, LOGL_DEBUG, "HNB-REGISTER-REQ %s MCC=%u,MNC=%u,LAC=%u,RAC=%u,SAC=%u,CID=%u from %s%s\n",
-	       ctx->identity_info, ctx->id.mcc, ctx->id.mnc, ctx->id.lac, ctx->id.rac, ctx->id.sac, ctx->id.cid,
-	       name, ctx->hnb_registered ? " (duplicated)" : "");
+	LOGHNB(ctx, DHNBAP, LOGL_DEBUG, "HNB-REGISTER-REQ %s %s%s\n",
+	       ctx->identity_info, umts_cell_id_name(&ctx->id), ctx->hnb_registered ? " (re-connecting)" : "");
+
+	/* The HNB is already registered, and we are seeing a new HNB Register Request. The HNB has restarted
+	 * without us noticing. Clearly, the HNB does not expect any UE state to be active here, so discard any
+	 * UE contexts and SCCP connections associated with this HNB. */
+	LOGHNB(ctx, DHNBAP, LOGL_NOTICE, "HNB (re)connecting, discarding all previous UE state\n");
+	hnb_context_release_ue_state(ctx);
 
 	ctx->hnb_registered = true;
 
@@ -504,28 +507,26 @@ static int hnbgw_rx_ue_register_req(struct hnb_context *ctx, ANY_t *in)
 		break;
 	case HNBAP_UE_Identity_PR_tMSILAI:
 	case HNBAP_UE_Identity_PR_pTMSIRAI:
-		if (ctx->gw->config.hnbap_allow_tmsi)
+		if (g_hnbgw->config.hnbap_allow_tmsi)
 			rc = hnbgw_tx_ue_register_acc_tmsi(ctx, &ies.uE_Identity);
 		else
 			rc = hnbgw_tx_ue_register_rej_tmsi(ctx, &ies.uE_Identity);
 		/* all has been handled by TMSI, skip the IMSI code below */
-		hnbap_free_ueregisterrequesties(&ies);
-		return rc;
+		goto free_and_return_rc;
 	default:
 		LOGHNB(ctx, DHNBAP, LOGL_NOTICE, "UE-REGISTER-REQ with unsupported UE Id type %d\n",
 			ies.uE_Identity.present);
-		hnbap_free_ueregisterrequesties(&ies);
-		return rc;
+		rc = -ENOTSUP;
+		goto free_and_return_rc;
 	}
 
 	LOGHNB(ctx, DHNBAP, LOGL_DEBUG, "UE-REGISTER-REQ ID_type=%d imsi=%s cause=%ld\n",
 		ies.uE_Identity.present, imsi, ies.registration_Cause);
 
-	ue = ue_context_by_imsi(ctx->gw, imsi);
+	ue = ue_context_by_imsi(imsi);
 	if (!ue)
 		ue = ue_allocated = ue_context_alloc(ctx, imsi, 0);
 
-	hnbap_free_ueregisterrequesties(&ies);
 	/* Send UERegisterAccept */
 	rc = hnbgw_tx_ue_register_acc(ue);
 	if (rc < 0) {
@@ -534,6 +535,8 @@ static int hnbgw_rx_ue_register_req(struct hnb_context *ctx, ANY_t *in)
 		if (ue_allocated)
 			ue_context_free(ue_allocated);
 	}
+free_and_return_rc:
+	hnbap_free_ueregisterrequesties(&ies);
 	return rc;
 }
 
@@ -552,7 +555,7 @@ static int hnbgw_rx_ue_deregister(struct hnb_context *ctx, ANY_t *in)
 
 	LOGHNB(ctx, DHNBAP, LOGL_DEBUG, "UE-DE-REGISTER context=%u cause=%s\n", ctxid, hnbap_cause_str(&ies.cause));
 
-	ue = ue_context_by_id(ctx->gw, ctxid);
+	ue = ue_context_by_id(ctxid);
 	if (ue)
 		ue_context_free(ue);
 
