@@ -33,35 +33,25 @@
 
 #include "asn1helpers.h"
 
-#include <osmocom/hnbgw/hnbgw.h>
+#include <osmocom/hnbgw/hnbgw_cn.h>
 #include <osmocom/hnbgw/hnbgw_ranap.h>
 #include <osmocom/rua/rua_common.h>
 #include <osmocom/rua/rua_ies_defs.h>
 #include <osmocom/hnbgw/context_map.h>
+#include <osmocom/hnbgw/hnbgw_rua.h>
 #include <osmocom/hnbap/HNBAP_CN-DomainIndicator.h>
-#include <osmocom/hnbgw/mgw_fsm.h>
-#include <osmocom/hnbgw/ps_rab_ass_fsm.h>
-#include <osmocom/ranap/RANAP_ProcedureCode.h>
-#include <osmocom/ranap/ranap_common.h>
-#include <osmocom/ranap/ranap_common_cn.h>
-
-static const char *cn_domain_indicator_to_str(RUA_CN_DomainIndicator_t cN_DomainIndicator)
-{
-	switch (cN_DomainIndicator) {
-	case RUA_CN_DomainIndicator_cs_domain:
-		return "IuCS";
-	case RUA_CN_DomainIndicator_ps_domain:
-		return "IuPS";
-	default:
-		return "(unknown-domain)";
-	}
-}
+#include <osmocom/ranap/ranap_ies_defs.h>
 
 static int hnbgw_rua_tx(struct hnb_context *ctx, struct msgb *msg)
 {
 	if (!msg)
 		return -EINVAL;
 
+	if (!ctx || !ctx->conn) {
+		LOGHNB(ctx, DRUA, LOGL_ERROR, "RUA context to this HNB is not connected, cannot transmit message\n");
+		return -ENOTCONN;
+	}
+
 	msgb_sctp_ppid(msg) = IUH_PPI_RUA;
 	osmo_stream_srv_send(ctx->conn, msg);
 
@@ -128,7 +118,7 @@ int rua_tx_dt(struct hnb_context *hnb, int is_ps, uint32_t context_id,
 					      &out);
 	ASN_STRUCT_FREE_CONTENTS_ONLY(asn_DEF_RUA_DirectTransfer, &out);
 
-	LOGHNB(hnb, DRUA, LOGL_DEBUG, "transmitting RUA (cn=%s) payload of %u bytes\n",
+	LOGHNB(hnb, DRUA, LOGL_DEBUG, "transmitting RUA DirectTransfer (cn=%s) payload of %u bytes\n",
 		is_ps ? "ps" : "cs", msgb_length(msg));
 
 	return hnbgw_rua_tx(hnb, msg);
@@ -156,8 +146,6 @@ int rua_tx_disc(struct hnb_context *hnb, int is_ps, uint32_t context_id,
 		ies.ranaP_Message.size = len;
 	}
 
-	/* FIXME: msgb_free(msg)? ownership not yet clear */
-
 	memset(&out, 0, sizeof(out));
 	rc = rua_encode_disconnecties(&out, &ies);
 	if (rc < 0)
@@ -169,39 +157,87 @@ int rua_tx_disc(struct hnb_context *hnb, int is_ps, uint32_t context_id,
 					      &out);
 	ASN_STRUCT_FREE_CONTENTS_ONLY(asn_DEF_RUA_Disconnect, &out);
 
-	LOGHNB(hnb, DRUA, LOGL_DEBUG, "transmitting RUA (cn=%s) payload of %u bytes\n",
-		is_ps ? "ps" : "cs", msgb_length(msg));
 
+	LOGHNB(hnb, DRUA, LOGL_DEBUG, "transmitting RUA Disconnect (cn=%s) payload of %u bytes\n",
+	       is_ps ? "ps" : "cs", msgb_length(msg));
 
 	return hnbgw_rua_tx(hnb, msg);
 }
 
+static struct value_string rua_procedure_code_names[] = {
+	{ RUA_ProcedureCode_id_Connect, "Connect" },
+	{ RUA_ProcedureCode_id_DirectTransfer, "DirectTransfer" },
+	{ RUA_ProcedureCode_id_Disconnect, "Disconnect" },
+	{ RUA_ProcedureCode_id_ConnectionlessTransfer, "ConnectionlessTransfer" },
+	{ RUA_ProcedureCode_id_ErrorIndication, "ErrorIndication" },
+	{ RUA_ProcedureCode_id_privateMessage, "PrivateMessage" },
+	{}
+};
 
-
-/* forward a RUA message to the SCCP User API to SCCP */
-int rua_to_scu(struct hnb_context *hnb,
-	       RUA_CN_DomainIndicator_t cN_DomainIndicator,
-	       enum osmo_scu_prim_type type,
-	       uint32_t context_id, uint32_t cause,
-	       const uint8_t *data, unsigned int len)
+static inline const char *rua_procedure_code_name(enum RUA_ProcedureCode val)
 {
-	struct msgb *msg;
-	struct osmo_scu_prim *prim;
+	return get_value_string(rua_procedure_code_names, val);
+}
+
+static struct hnbgw_context_map *find_or_create_context_map(struct hnb_context *hnb, uint32_t rua_ctx_id, bool is_ps,
+							    struct msgb *ranap_msg)
+{
+	struct hnbgw_context_map *map;
+	struct hnbgw_cnlink *cnlink;
+
+	map = context_map_find_by_rua_ctx_id(hnb, rua_ctx_id, is_ps);
+	if (map) {
+		/* Already established SCCP link. Continue to use that. */
+		return map;
+	}
+
+	/* Establish a new context map. From the RUA Connect, extract a mobile identity, if any, and select a CN link
+	 * based on an NRI found in the mobile identity, if any. */
+
+	/* Allocate a map for logging context */
+	map = context_map_alloc(hnb, rua_ctx_id, is_ps);
+	OSMO_ASSERT(map);
+
+	if (hnbgw_peek_l3(map, ranap_msg))
+		LOGP(DCN, LOGL_NOTICE, "Failed to extract Mobile Identity from RUA Connect message's RANAP payload\n");
+	/* map->l3 now contains all the interesting information from the NAS PDU, if any.
+	 * If no useful information could be decoded, still continue to select a hopefully adequate link by round robin.
+	 */
+
+	cnlink = hnbgw_cnlink_select(map);
+	if (!cnlink) {
+		LOG_MAP(map, DCN, LOGL_ERROR, "Failed to select %s link\n", is_ps ? "IuPS" : "IuCS");
+		context_map_free(map);
+		return NULL;
+	}
+
+	if (context_map_set_cnlink(map, cnlink)) {
+		LOG_MAP(map, DCN, LOGL_ERROR, "Failed to establish link to %s\n", cnlink->name);
+		context_map_free(map);
+		return NULL;
+	}
+
+	return map;
+}
+
+/* dispatch a RUA connection-oriented message received from a HNB to a context mapping's RUA FSM, so that it is
+ * forwarded to the CN via SCCP connection-oriented messages.
+ * Connectionless messages are handled in hnbgw_ranap_rx() instead, not here. */
+static int rua_to_scu(struct hnb_context *hnb,
+		      RUA_CN_DomainIndicator_t cN_DomainIndicator,
+		      enum RUA_ProcedureCode rua_procedure,
+		      uint32_t context_id, uint32_t cause,
+		      const uint8_t *data, unsigned int len)
+{
+	struct msgb *ranap_msg = NULL;
 	struct hnbgw_context_map *map = NULL;
-	struct hnbgw_cnlink *cn = hnb->gw->sccp.cnlink;
-	struct osmo_sccp_addr *remote_addr;
 	bool is_ps;
-	bool release_context_map = false;
-	ranap_message *message;
-	int rc;
 
 	switch (cN_DomainIndicator) {
 	case RUA_CN_DomainIndicator_cs_domain:
-		remote_addr = &hnb->gw->sccp.iucs_remote_addr;
 		is_ps = false;
 		break;
 	case RUA_CN_DomainIndicator_ps_domain:
-		remote_addr = &hnb->gw->sccp.iups_remote_addr;
 		is_ps = true;
 		break;
 	default:
@@ -209,115 +245,42 @@ int rua_to_scu(struct hnb_context *hnb,
 		return -1;
 	}
 
-	if (!cn) {
-		LOGHNB(hnb, DRUA, LOGL_NOTICE, "CN=NULL, discarding message\n");
-		return 0;
-	}
-
-	msg = msgb_alloc(1500, "rua_to_sccp");
-
-	prim = (struct osmo_scu_prim *) msgb_put(msg, sizeof(*prim));
-	osmo_prim_init(&prim->oph, SCCP_SAP_USER, type, PRIM_OP_REQUEST, msg);
-
-	switch (type) {
-	case OSMO_SCU_PRIM_N_UNITDATA:
-		LOGHNB(hnb, DRUA, LOGL_DEBUG, "rua_to_scu() %s to %s, rua_ctx_id %u (unitdata, no scu_conn_id)\n",
-			cn_domain_indicator_to_str(cN_DomainIndicator), osmo_sccp_addr_dump(remote_addr), context_id);
-		break;
-	default:
-		map = context_map_alloc_by_hnb(hnb, context_id, is_ps, cn);
-		OSMO_ASSERT(map);
-		LOGHNB(hnb, DRUA, LOGL_DEBUG, "rua_to_scu() %s to %s, rua_ctx_id %u scu_conn_id %u data-len %u\n",
-		       cn_domain_indicator_to_str(cN_DomainIndicator), osmo_sccp_addr_dump(remote_addr),
-		       map->rua_ctx_id, map->scu_conn_id, len);
-	}
-
-	/* add primitive header */
-	switch (type) {
-	case OSMO_SCU_PRIM_N_CONNECT:
-		prim->u.connect.called_addr = *remote_addr;
-		prim->u.connect.calling_addr = cn->gw->sccp.local_addr;
-		prim->u.connect.sccp_class = 2;
-		prim->u.connect.conn_id = map->scu_conn_id;
-		/* Two separate logs because of osmo_sccp_addr_dump(). */
-		LOGHNB(hnb, DRUA, LOGL_DEBUG, "RUA to SCCP N_CONNECT: called_addr:%s\n",
-			osmo_sccp_addr_dump(&prim->u.connect.called_addr));
-		LOGHNB(hnb, DRUA, LOGL_DEBUG, "RUA to SCCP N_CONNECT: calling_addr:%s\n",
-			osmo_sccp_addr_dump(&prim->u.connect.calling_addr));
-		break;
-	case OSMO_SCU_PRIM_N_DATA:
-		prim->u.data.conn_id = map->scu_conn_id;
-		break;
-	case OSMO_SCU_PRIM_N_DISCONNECT:
-		prim->u.disconnect.conn_id = map->scu_conn_id;
-		prim->u.disconnect.cause = cause;
-		release_context_map = true;
-		/* Mark SCCP conn as gracefully disconnected */
-		map->scu_conn_active = false;
-		break;
-	case OSMO_SCU_PRIM_N_UNITDATA:
-		prim->u.unitdata.called_addr = *remote_addr;
-		prim->u.unitdata.calling_addr = cn->gw->sccp.local_addr;
-		/* Two separate logs because of osmo_sccp_addr_dump(). */
-		LOGHNB(hnb, DRUA, LOGL_DEBUG, "RUA to SCCP N_UNITDATA: called_addr:%s\n",
-			osmo_sccp_addr_dump(&prim->u.unitdata.called_addr));
-		LOGHNB(hnb, DRUA, LOGL_DEBUG, "RUA to SCCP N_UNITDATA: calling_addr:%s\n",
-			osmo_sccp_addr_dump(&prim->u.unitdata.calling_addr));
-		break;
-	default:
-		return -EINVAL;
-	}
-
-	/* If there is RANAP data, include it in the msgb. Usually there is data, but this could also be an SCCP CR
-	 * a.k.a. OSMO_SCU_PRIM_N_CONNECT without RANAP payload. */
+	/* If there is RANAP data, include it in the msgb. In RUA there is always data in practice, but theoretically it
+	 * could be an empty Connect or Disconnect. */
 	if (data && len) {
-		msg->l2h = msgb_put(msg, len);
-		memcpy(msg->l2h, data, len);
+		/* According to API doc of map_rua_fsm_event: allocate msgb for RANAP data from OTC_SELECT, reserve
+		 * headroom for an osmo_scu_prim. Point l2h at the RANAP data. */
+		ranap_msg = hnbgw_ranap_msg_alloc("RANAP_from_RUA");
+		ranap_msg->l2h = msgb_put(ranap_msg, len);
+		memcpy(ranap_msg->l2h, data, len);
 	}
 
-	/* If there is data, see if it is a RAB Assignment message where we need to change the user plane information,
-	 * for RTP mapping via MGW (soon also GTP mapping via UPF). */
-	if (data && len && map && !release_context_map) {
-		if (!map->is_ps) {
-			message = talloc_zero(map, ranap_message);
-			rc = ranap_cn_rx_co_decode(map, message, msgb_l2(prim->oph.msg), msgb_l2len(prim->oph.msg));
-
-			if (rc == 0) {
-				switch (message->procedureCode) {
-				case RANAP_ProcedureCode_id_RAB_Assignment:
-					/* mgw_fsm_handle_rab_ass_resp() takes ownership of prim->oph and (ranap) message */
-					return mgw_fsm_handle_rab_ass_resp(map, &prim->oph, message);
-				}
-				ranap_cn_rx_co_free(message);
-			}
-
-			talloc_free(message);
-#if ENABLE_PFCP
-		} else if (hnb_gw_is_gtp_mapping_enabled(hnb->gw)) {
-			/* map->is_ps == true and PFCP is enabled in osmo-hnbgw.cfg */
-			message = talloc_zero(map, ranap_message);
-			rc = ranap_cn_rx_co_decode(map, message, msgb_l2(prim->oph.msg), msgb_l2len(prim->oph.msg));
-
-			if (rc == 0) {
-				switch (message->procedureCode) {
-				case RANAP_ProcedureCode_id_RAB_Assignment:
-					/* ps_rab_ass_fsm takes ownership of prim->oph and RANAP message */
-					return hnbgw_gtpmap_rx_rab_ass_resp(map, &prim->oph, message);
-				}
-				ranap_cn_rx_co_free(message);
-			}
-
-			talloc_free(message);
-#endif
-		}
+	map = find_or_create_context_map(hnb, context_id, is_ps, ranap_msg);
+	if (!map) {
+		LOGHNB(hnb, DRUA, LOGL_ERROR,
+		       "Failed to create context map for %s: rx RUA %s with %u bytes RANAP data\n",
+		       is_ps ? "IuPS" : "IuCS", rua_procedure_code_name(rua_procedure), data ? len : 0);
+		return -1;
 	}
 
-	rc = osmo_sccp_user_sap_down(cn->sccp_user, &prim->oph);
+	LOG_MAP(map, DRUA, LOGL_DEBUG, "rx RUA %s with %u bytes RANAP data\n",
+		rua_procedure_code_name(rua_procedure), data ? len : 0);
 
-	if (map && release_context_map)
-		context_map_deactivate(map);
+	switch (rua_procedure) {
 
-	return rc;
+	case RUA_ProcedureCode_id_Connect:
+		return map_rua_dispatch(map, MAP_RUA_EV_RX_CONNECT, ranap_msg);
+
+	case RUA_ProcedureCode_id_DirectTransfer:
+		return map_rua_dispatch(map, MAP_RUA_EV_RX_DIRECT_TRANSFER, ranap_msg);
+
+	case RUA_ProcedureCode_id_Disconnect:
+		return map_rua_dispatch(map, MAP_RUA_EV_RX_DISCONNECT, ranap_msg);
+
+	default:
+		/* No caller may ever pass a different RUA procedure code */
+		OSMO_ASSERT(false);
+	}
 }
 
 static uint32_t rua_to_scu_cause(RUA_Cause_t *in)
@@ -388,69 +351,21 @@ static int rua_rx_init_connect(struct msgb *msg, ANY_t *in)
 	struct hnb_context *hnb = msg->dst;
 	uint32_t context_id;
 	int rc;
-	const uint8_t *data;
-	unsigned int data_len;
 
 	rc = rua_decode_connecties(&ies, in);
 	if (rc < 0)
 		return rc;
 
 	context_id = asn1bitstr_to_u24(&ies.context_ID);
-	data = ies.ranaP_Message.buf;
-	data_len = ies.ranaP_Message.size;
 
-	LOGHNB(hnb, DRUA, LOGL_DEBUG, "RUA %s Connect.req(ctx=0x%x, %s, RANAP.size=%u)\n",
-	       cn_domain_indicator_to_str(ies.cN_DomainIndicator), context_id,
-	       ies.establishment_Cause == RUA_Establishment_Cause_emergency_call ? "emergency" : "normal",
-	       data_len);
+	LOGHNB(hnb, DRUA, LOGL_DEBUG, "RUA %s Connect.req(ctx=0x%x, %s)\n",
+		ranap_domain_name(ies.cN_DomainIndicator), context_id,
+		ies.establishment_Cause == RUA_Establishment_Cause_emergency_call ? "emergency" : "normal");
 
-	if (hnbgw_requires_empty_sccp_cr(hnb->gw, data_len)) {
-		/* Do not include data in the SCCP CR, to avoid hitting a message size limit at the remote end that may
-		 * lead to rejection. */
-		bool is_ps;
-		struct osmo_sccp_addr *remote_addr;
-		struct hnbgw_context_map *map;
+	rc = rua_to_scu(hnb, ies.cN_DomainIndicator, RUA_ProcedureCode_id_Connect,
+			context_id, 0, ies.ranaP_Message.buf,
+			ies.ranaP_Message.size);
 
-		switch (ies.cN_DomainIndicator) {
-		case RUA_CN_DomainIndicator_cs_domain:
-			remote_addr = &hnb->gw->sccp.iucs_remote_addr;
-			is_ps = false;
-			break;
-		case RUA_CN_DomainIndicator_ps_domain:
-			remote_addr = &hnb->gw->sccp.iups_remote_addr;
-			is_ps = true;
-			break;
-		default:
-			LOGHNB(hnb, DRUA, LOGL_ERROR, "Unsupported Domain %ld\n", ies.cN_DomainIndicator);
-			rua_free_connecties(&ies);
-			return -1;
-		}
-
-		if (!hnb->gw->sccp.cnlink) {
-			LOGHNB(hnb, DRUA, LOGL_NOTICE, "CN=NULL, discarding message\n");
-			rua_free_connecties(&ies);
-			return 0;
-		}
-
-		map = context_map_alloc_by_hnb(hnb, context_id, is_ps, hnb->gw->sccp.cnlink);
-		OSMO_ASSERT(map);
-		OSMO_ASSERT(map->is_ps == is_ps);
-		LOGHNB(hnb, DRUA, LOGL_DEBUG, "rua_rx_init_connect() %s to %s, rua_ctx_id %u scu_conn_id %u;"
-		       " Sending SCCP CR without payload, caching %u octets\n",
-		       cn_domain_indicator_to_str(ies.cN_DomainIndicator), osmo_sccp_addr_dump(remote_addr),
-		       map->rua_ctx_id, map->scu_conn_id, data_len);
-
-		map->cached_msg = msgb_alloc_c(map, data_len, "map.cached_msg");
-		OSMO_ASSERT(map->cached_msg);
-		memcpy(msgb_put(map->cached_msg, data_len), data, data_len);
-
-		/* Data is cached for after CR is confirmed, send SCCP CR but omit payload. */
-		data = NULL;
-		data_len = 0;
-	}
-
-	rc = rua_to_scu(hnb, ies.cN_DomainIndicator, OSMO_SCU_PRIM_N_CONNECT,
-			context_id, 0, data, data_len);
 	rua_free_connecties(&ies);
 
 	return rc;
@@ -482,7 +397,7 @@ static int rua_rx_init_disconnect(struct msgb *msg, ANY_t *in)
 	}
 
 	rc = rua_to_scu(hnb, ies.cN_DomainIndicator,
-			OSMO_SCU_PRIM_N_DISCONNECT,
+			RUA_ProcedureCode_id_Disconnect,
 			context_id, scu_cause, ranap_data, ranap_len);
 
 	rua_free_disconnecties(&ies);
@@ -507,7 +422,7 @@ static int rua_rx_init_dt(struct msgb *msg, ANY_t *in)
 
 	rc = rua_to_scu(hnb,
 			ies.cN_DomainIndicator,
-			OSMO_SCU_PRIM_N_DATA,
+			RUA_ProcedureCode_id_DirectTransfer,
 			context_id, 0, ies.ranaP_Message.buf,
 			ies.ranaP_Message.size);
 
@@ -651,11 +566,7 @@ int hnbgw_rua_rx(struct hnb_context *hnb, struct msgb *msg)
 
 	rc = _hnbgw_rua_rx(msg, pdu);
 
+	ASN_STRUCT_FREE_CONTENTS_ONLY(asn_DEF_RUA_RUA_PDU, pdu);
+
 	return rc;
 }
-
-
-int hnbgw_rua_init(void)
-{
-	return 0;
-}
