@@ -52,10 +52,7 @@
 /* Send Iu Release Request, this is done in erroneous cases from which we cannot recover */
 static void tx_release_req(struct hnbgw_context_map *map)
 {
-	struct hnb_context *hnb = map->hnb_ctx;
-	struct hnbgw_cnlink *cn = hnb->gw->sccp.cnlink;
 	struct msgb *msg;
-	struct osmo_scu_prim *prim;
 	static const struct RANAP_Cause cause = {
 		.present = RANAP_Cause_PR_transmissionNetwork,
 		.choice.transmissionNetwork =
@@ -64,11 +61,8 @@ static void tx_release_req(struct hnbgw_context_map *map)
 
 	msg = ranap_new_msg_iu_rel_req(&cause);
 	msg->l2h = msg->data;
-
-	prim = (struct osmo_scu_prim *)msgb_push(msg, sizeof(*prim));
-	prim->u.data.conn_id = map->scu_conn_id;
-	osmo_prim_init(&prim->oph, SCCP_SAP_USER, OSMO_SCU_PRIM_N_DATA, PRIM_OP_REQUEST, msg);
-	osmo_sccp_user_sap_down(cn->sccp_user, &prim->oph);
+	talloc_steal(OTC_SELECT, msg);
+	map_sccp_dispatch(map, MAP_SCCP_EV_TX_DATA_REQUEST, msg);
 }
 
 #define S(x)	(1 << (x))
@@ -112,13 +106,14 @@ struct mgw_fsm_priv {
 	/* Pointers to messages and prim header we take ownership of */
 	ranap_message *ranap_rab_ass_req_message;
 	ranap_message *ranap_rab_ass_resp_message;
-	struct osmo_prim_hdr *ranap_rab_ass_resp_oph;
+	struct msgb *ranap_rab_ass_resp_msgb;
 
 	/* MGW context */
 	struct mgcp_client *mgcpc;
 	struct osmo_mgcpc_ep *mgcpc_ep;
 	struct osmo_mgcpc_ep_ci *mgcpc_ep_ci_hnb;
 	struct osmo_mgcpc_ep_ci *mgcpc_ep_ci_msc;
+	struct osmo_sockaddr ci_hnb_crcx_ack_addr;
 	char msc_rtp_addr[INET6_ADDRSTRLEN];
 	uint16_t msc_rtp_port;
 };
@@ -173,7 +168,7 @@ static void mgw_fsm_crcx_hnb_onenter(struct osmo_fsm_inst *fi, uint32_t prev_sta
 	mgw_info.codecs[0] = CODEC_IUFP;
 	mgw_info.codecs_len = 1;
 
-	mgw_fsm_priv->mgcpc = mgcp_client_pool_get(map->hnb_ctx->gw->mgw_pool);
+	mgw_fsm_priv->mgcpc = mgcp_client_pool_get(g_hnbgw->mgw_pool);
 	if (!mgw_fsm_priv->mgcpc) {
 		LOGPFSML(fi, LOGL_ERROR,
 			 "cannot ensure MGW endpoint -- no MGW configured, check configuration!\n");
@@ -193,8 +188,8 @@ static void mgw_fsm_crcx_hnb(struct osmo_fsm_inst *fi, uint32_t event, void *dat
 {
 	struct mgw_fsm_priv *mgw_fsm_priv = fi->priv;
 	const struct mgcp_conn_peer *mgw_info;
-	struct osmo_sockaddr addr;
 	struct osmo_sockaddr_str addr_str;
+	struct osmo_sockaddr *addr = &mgw_fsm_priv->ci_hnb_crcx_ack_addr;
 	RANAP_RAB_AssignmentRequestIEs_t *ies;
 	int rc;
 
@@ -213,7 +208,7 @@ static void mgw_fsm_crcx_hnb(struct osmo_fsm_inst *fi, uint32_t event, void *dat
 			addr_str.af = AF_INET6;
 		addr_str.port = mgw_info->port;
 		osmo_strlcpy(addr_str.ip, mgw_info->addr, sizeof(addr_str.ip));
-		rc = osmo_sockaddr_str_to_sockaddr(&addr_str, &addr.u.sas);
+		rc = osmo_sockaddr_str_to_sockaddr(&addr_str, &addr->u.sas);
 		if (rc < 0) {
 			LOGPFSML(fi, LOGL_ERROR,
 				 "Failed to convert RTP IP-address (%s) and Port (%u) to its binary representation\n",
@@ -223,7 +218,7 @@ static void mgw_fsm_crcx_hnb(struct osmo_fsm_inst *fi, uint32_t event, void *dat
 		}
 
 		ies = &mgw_fsm_priv->ranap_rab_ass_req_message->msg.raB_AssignmentRequestIEs;
-		rc = ranap_rab_ass_req_ies_replace_inet_addr(ies, &addr, mgw_fsm_priv->rab_id);
+		rc = ranap_rab_ass_req_ies_replace_inet_addr(ies, addr, mgw_fsm_priv->rab_id);
 		if (rc < 0) {
 			LOGPFSML(fi, LOGL_ERROR,
 				 "Failed to replace RTP IP-address (%s) and Port (%u) in RAB-AssignmentRequest\n",
@@ -255,8 +250,9 @@ static void mgw_fsm_assign_onenter(struct osmo_fsm_inst *fi, uint32_t prev_state
 	}
 
 	LOGPFSML(fi, LOGL_DEBUG, "forwarding modified RAB-AssignmentRequest to HNB\n");
-	rua_tx_dt(map->hnb_ctx, map->is_ps, map->rua_ctx_id, msg->data, msg->len);
-	msgb_free(msg);
+	msg->l2h = msg->data;
+	talloc_steal(OTC_SELECT, msg);
+	map_rua_dispatch(map, MAP_RUA_EV_TX_DIRECT_TRANSFER, msg);
 }
 
 static void mgw_fsm_assign(struct osmo_fsm_inst *fi, uint32_t event, void *data)
@@ -274,8 +270,6 @@ static void mgw_fsm_mdcx_hnb_onenter(struct osmo_fsm_inst *fi, uint32_t prev_sta
 {
 	struct mgw_fsm_priv *mgw_fsm_priv = fi->priv;
 	struct hnbgw_context_map *map = mgw_fsm_priv->map;
-	struct hnb_context *hnb = map->hnb_ctx;
-	struct hnbgw_cnlink *cn = hnb->gw->sccp.cnlink;
 	struct mgcp_conn_peer mgw_info;
 	struct osmo_sockaddr addr;
 	struct osmo_sockaddr_str addr_str;
@@ -298,6 +292,8 @@ static void mgw_fsm_mdcx_hnb_onenter(struct osmo_fsm_inst *fi, uint32_t prev_sta
 	if (rc < 0) {
 		rab_failed_at_hnb = ranap_rab_ass_resp_ies_check_failure(ies, mgw_fsm_priv->rab_id);
 		if (rab_failed_at_hnb) {
+			struct msgb *msg;
+
 			LOGPFSML(fi, LOGL_ERROR,
 				 "The RAB-AssignmentResponse contains a RAB-FailedList, RAB-Assignment (%u) failed.\n",
 				 mgw_fsm_priv->rab_id);
@@ -305,8 +301,12 @@ static void mgw_fsm_mdcx_hnb_onenter(struct osmo_fsm_inst *fi, uint32_t prev_sta
 			/* Forward the RAB-AssignmentResponse transparently. This will ensure that the MSC is informed
 			 * about the problem. */
 			LOGPFSML(fi, LOGL_DEBUG, "forwarding unmodified RAB-AssignmentResponse to MSC\n");
-			rc = osmo_sccp_user_sap_down(cn->sccp_user, mgw_fsm_priv->ranap_rab_ass_resp_oph);
-			mgw_fsm_priv->ranap_rab_ass_resp_oph = NULL;
+
+			msg = mgw_fsm_priv->ranap_rab_ass_resp_msgb;
+			mgw_fsm_priv->ranap_rab_ass_resp_msgb = NULL;
+			talloc_steal(OTC_SELECT, msg);
+
+			rc = map_sccp_dispatch(map, MAP_SCCP_EV_TX_DATA_REQUEST, msg);
 			if (rc < 0) {
 				LOGPFSML(fi, LOGL_DEBUG, "failed to forward RAB-AssignmentResponse message\n");
 				osmo_fsm_inst_state_chg(fi, MGW_ST_FAILURE, 0, 0);
@@ -342,6 +342,9 @@ static void mgw_fsm_mdcx_hnb(struct osmo_fsm_inst *fi, uint32_t event, void *dat
 {
 	struct mgw_fsm_priv *mgw_fsm_priv = fi->priv;
 	const struct mgcp_conn_peer *mgw_info;
+	struct osmo_sockaddr_str addr_str;
+	struct osmo_sockaddr addr;
+	int rc;
 
 	switch (event) {
 	case MGW_EV_MGCP_OK:
@@ -351,6 +354,33 @@ static void mgw_fsm_mdcx_hnb(struct osmo_fsm_inst *fi, uint32_t event, void *dat
 			osmo_fsm_inst_state_chg(fi, MGW_ST_FAILURE, 0, 0);
 			return;
 		}
+
+		if (strchr(mgw_info->addr, '.'))
+			addr_str.af = AF_INET;
+		else
+			addr_str.af = AF_INET6;
+		addr_str.port = mgw_info->port;
+		osmo_strlcpy(addr_str.ip, mgw_info->addr, sizeof(addr_str.ip));
+		rc = osmo_sockaddr_str_to_sockaddr(&addr_str, &addr.u.sas);
+		if (rc < 0) {
+			LOGPFSML(fi, LOGL_ERROR,
+				 "Failed to convert RTP IP-address (%s) and Port (%u) to its binary representation\n",
+				 mgw_info->addr, mgw_info->port);
+			osmo_fsm_inst_state_chg(fi, MGW_ST_FAILURE, 0, 0);
+			return;
+		}
+
+		if (osmo_sockaddr_cmp(&mgw_fsm_priv->ci_hnb_crcx_ack_addr, &addr) != 0) {
+			/* FIXME: Send RAB Modify Req to HNB. See OS#6127 */
+			char addr_buf[INET6_ADDRSTRLEN + 8];
+			LOGPFSML(fi, LOGL_ERROR, "Local MGW IuUP IP address %s changed to %s during MDCX."
+				 " This is so far unsupported, adapt your osmo-mgw config!\n",
+				 osmo_sockaddr_to_str(&mgw_fsm_priv->ci_hnb_crcx_ack_addr),
+				 osmo_sockaddr_to_str_buf(addr_buf, sizeof(addr_buf), &addr));
+			osmo_fsm_inst_state_chg(fi, MGW_ST_FAILURE, 0, 0);
+			return;
+		}
+
 		mgw_fsm_state_chg(fi, MGW_ST_CRCX_MSC);
 		return;
 	default:
@@ -431,15 +461,15 @@ static void mgw_fsm_crcx_msc(struct osmo_fsm_inst *fi, uint32_t event, void *dat
 		 * the original message. Ensure that there is enough room in l2h to grow. (The current implementation
 		 * should yield a message with the same size, but there is no guarantee for that) */
 		msg_max_len =
-		    msgb_l2len(mgw_fsm_priv->ranap_rab_ass_resp_oph->msg) +
-		    msgb_tailroom(mgw_fsm_priv->ranap_rab_ass_resp_oph->msg);
-		rc = msgb_resize_area(mgw_fsm_priv->ranap_rab_ass_resp_oph->msg,
-				      mgw_fsm_priv->ranap_rab_ass_resp_oph->msg->l2h,
-				      msgb_l2len(mgw_fsm_priv->ranap_rab_ass_resp_oph->msg), msg_max_len);
+		    msgb_l2len(mgw_fsm_priv->ranap_rab_ass_resp_msgb) +
+		    msgb_tailroom(mgw_fsm_priv->ranap_rab_ass_resp_msgb);
+		rc = msgb_resize_area(mgw_fsm_priv->ranap_rab_ass_resp_msgb,
+				      mgw_fsm_priv->ranap_rab_ass_resp_msgb->l2h,
+				      msgb_l2len(mgw_fsm_priv->ranap_rab_ass_resp_msgb), msg_max_len);
 		OSMO_ASSERT(rc == 0);
 
-		rc = ranap_rab_ass_resp_encode(msgb_l2(mgw_fsm_priv->ranap_rab_ass_resp_oph->msg),
-					       msgb_l2len(mgw_fsm_priv->ranap_rab_ass_resp_oph->msg), ies);
+		rc = ranap_rab_ass_resp_encode(msgb_l2(mgw_fsm_priv->ranap_rab_ass_resp_msgb),
+					       msgb_l2len(mgw_fsm_priv->ranap_rab_ass_resp_msgb), ies);
 		if (rc < 0) {
 			LOGPFSML(fi, LOGL_ERROR, "failed to re-encode RAB-AssignmentResponse message\n");
 			osmo_fsm_inst_state_chg(fi, MGW_ST_FAILURE, 0, 0);
@@ -447,9 +477,9 @@ static void mgw_fsm_crcx_msc(struct osmo_fsm_inst *fi, uint32_t event, void *dat
 		}
 
 		/* Resize l2h back to the actual message length */
-		rc = msgb_resize_area(mgw_fsm_priv->ranap_rab_ass_resp_oph->msg,
-				      mgw_fsm_priv->ranap_rab_ass_resp_oph->msg->l2h,
-				      msgb_l2len(mgw_fsm_priv->ranap_rab_ass_resp_oph->msg), rc);
+		rc = msgb_resize_area(mgw_fsm_priv->ranap_rab_ass_resp_msgb,
+				      mgw_fsm_priv->ranap_rab_ass_resp_msgb->l2h,
+				      msgb_l2len(mgw_fsm_priv->ranap_rab_ass_resp_msgb), rc);
 		OSMO_ASSERT(rc == 0);
 
 		/* When the established state is entered, the modified RAB AssignmentResponse is forwarded to the MSC.
@@ -466,14 +496,16 @@ static void mgw_fsm_established_onenter(struct osmo_fsm_inst *fi, uint32_t prev_
 {
 	struct mgw_fsm_priv *mgw_fsm_priv = fi->priv;
 	struct hnbgw_context_map *map = mgw_fsm_priv->map;
-	struct osmo_prim_hdr *oph = mgw_fsm_priv->ranap_rab_ass_resp_oph;
-	struct hnb_context *hnb = map->hnb_ctx;
-	struct hnbgw_cnlink *cn = hnb->gw->sccp.cnlink;
+	struct msgb *ranap_msg;
 	int rc;
 
 	LOGPFSML(fi, LOGL_DEBUG, "forwarding modified RAB-AssignmentResponse to MSC\n");
-	rc = osmo_sccp_user_sap_down(cn->sccp_user, oph);
-	mgw_fsm_priv->ranap_rab_ass_resp_oph = NULL;
+
+	ranap_msg = mgw_fsm_priv->ranap_rab_ass_resp_msgb;
+	mgw_fsm_priv->ranap_rab_ass_resp_msgb = NULL;
+	talloc_steal(OTC_SELECT, ranap_msg);
+
+	rc = map_sccp_dispatch(map, MAP_SCCP_EV_TX_DATA_REQUEST, ranap_msg);
 	if (rc < 0) {
 		LOGPFSML(fi, LOGL_DEBUG, "failed to forward RAB-AssignmentResponse message\n");
 		osmo_fsm_inst_state_chg(fi, MGW_ST_FAILURE, 0, 0);
@@ -530,28 +562,6 @@ static int mgw_fsm_timer_cb(struct osmo_fsm_inst *fi)
 static void mgw_fsm_cleanup(struct osmo_fsm_inst *fi, enum osmo_fsm_term_cause cause)
 {
 	struct mgw_fsm_priv *mgw_fsm_priv = fi->priv;
-	struct osmo_scu_prim *scu_prim;
-	struct msgb *scu_msg;
-
-	if (mgw_fsm_priv->ranap_rab_ass_req_message) {
-		ranap_ran_rx_co_free(mgw_fsm_priv->ranap_rab_ass_req_message);
-		talloc_free(mgw_fsm_priv->ranap_rab_ass_req_message);
-		mgw_fsm_priv->ranap_rab_ass_req_message = NULL;
-	}
-
-	if (mgw_fsm_priv->ranap_rab_ass_resp_message) {
-		ranap_cn_rx_co_free(mgw_fsm_priv->ranap_rab_ass_resp_message);
-		talloc_free(mgw_fsm_priv->ranap_rab_ass_resp_message);
-		mgw_fsm_priv->ranap_rab_ass_resp_message = NULL;
-	}
-
-	if (mgw_fsm_priv->ranap_rab_ass_resp_oph) {
-		scu_prim = (struct osmo_scu_prim *)mgw_fsm_priv->ranap_rab_ass_resp_oph;
-		scu_msg = scu_prim->oph.msg;
-		msgb_free(scu_msg);
-		mgw_fsm_priv->ranap_rab_ass_resp_oph = NULL;
-	}
-
 	talloc_free(mgw_fsm_priv);
 }
 
@@ -654,43 +664,56 @@ static struct osmo_fsm mgw_fsm = {
 };
 
 /* The MSC may ask to release a specific RAB within a RAB-AssignmentRequest */
-static int handle_rab_release(struct hnbgw_context_map *map, struct osmo_prim_hdr *oph, ranap_message *message)
+static int release_mgw_fsm(struct hnbgw_context_map *map, struct msgb *ranap_msg)
 {
-	bool rab_release_req;
 	struct osmo_fsm_inst *fi = map->mgw_fi;
-	struct mgw_fsm_priv *mgw_fsm_priv = fi->priv;
 	int rc;
 
-	/* Check if the RAB that is handled by this FSM is addressed by the release request */
-	rab_release_req = ranap_rab_ass_req_ies_check_release(&message->msg.raB_AssignmentRequestIEs,
-							      mgw_fsm_priv->rab_id);
-	if (!rab_release_req)
-		return -EINVAL;
-
-	LOGPFSML(map->mgw_fi, LOGL_NOTICE, "MSC asked to release RAB-ID %u\n", mgw_fsm_priv->rab_id);
-
 	/* Forward the unmodifed RAB-AssignmentRequest to HNB, so that the HNB is informed about the RAB release as
 	 * well */
 	LOGPFSML(fi, LOGL_DEBUG, "forwarding unmodified RAB-AssignmentRequest to HNB\n");
-	rc = rua_tx_dt(map->hnb_ctx, map->is_ps, map->rua_ctx_id, msgb_l2(oph->msg), msgb_l2len(oph->msg));
+	rc = map_rua_dispatch(map, MAP_RUA_EV_TX_DIRECT_TRANSFER, ranap_msg);
+	if (rc < 0) {
+		LOGPFSML(fi, LOGL_DEBUG, "cannot forward RAB-AssignmentRequest to HNB\n");
+		return -EINVAL;
+	}
 
 	/* Release the FSM normally */
 	osmo_fsm_inst_state_chg(fi, MGW_ST_RELEASE, 0, 0);
-
-	return rc;
+	return 0;
 }
 
-/*! Allocate MGW FSM and handle RANAP RAB AssignmentRequest).
- *  \ptmap[in] map hanbgw context map that is responsible for this call.
- *  \ptmap[in] oph osmo prim header with RANAP RAB AssignmentResponse (function takes no ownership).
- *  \ptmap[in] message ranap message container (function takes ownership).
+/* Check if the message contains a RAB-ReleaseItem that matches the RAB-ID that is managed by the given context map */
+static bool is_our_rab_release(struct hnbgw_context_map *map, ranap_message *message)
+{
+	bool rab_release_req;
+	struct osmo_fsm_inst *fi = map->mgw_fi;
+	struct mgw_fsm_priv *mgw_fsm_priv = fi->priv;
+
+	/* Check if the RAB that is handled by this FSM is addressed by the release request */
+	rab_release_req = ranap_rab_ass_req_ies_check_release(&message->msg.raB_AssignmentRequestIEs,
+							      mgw_fsm_priv->rab_id);
+	if (!rab_release_req) {
+		LOGPFSML(map->mgw_fi, LOGL_ERROR, "RAB-AssignmentRequest does not contain any RAB-RelaseItem with RAB-ID %u\n", mgw_fsm_priv->rab_id);
+		return false;
+	}
+	LOGPFSML(map->mgw_fi, LOGL_NOTICE, "MSC asked to release RAB-ID %u\n", mgw_fsm_priv->rab_id);
+
+	return true;
+}
+
+/*! Allocate MGW FSM and handle RANAP RAB AssignmentRequest.
+ *  \param[in] map hnbgw context map that is responsible for this call.
+ *  \param[in] ranap_msg msgb containing RANAP RAB AssignmentRequest at msgb_l2(), allocated in OTC_SELECT.
+ *                       This function may talloc_steal(ranap_msg) to keep it for later.
+ *  \param[in] message decoded RANAP message container, allocated in OTC_SELECT.
+ *                     This function may talloc_steal(message) to keep it for later.
  *  \returns 0 on success; negative on error. */
-int handle_rab_ass_req(struct hnbgw_context_map *map, struct osmo_prim_hdr *oph, ranap_message *message)
+int handle_rab_ass_req(struct hnbgw_context_map *map, struct msgb *ranap_msg, ranap_message *message)
 {
 	static bool initialized = false;
 	struct mgw_fsm_priv *mgw_fsm_priv;
 	char fsm_name[255];
-	int rc;
 
 	/* Initialize FSM if not done yet */
 	if (!initialized) {
@@ -699,16 +722,19 @@ int handle_rab_ass_req(struct hnbgw_context_map *map, struct osmo_prim_hdr *oph,
 	}
 
 	/* The RTP stream negotiation usually begins with a RAB-AssignmentRequest and ends with an IU-Release, however
-	 * it may also be thet the MSC decides to release the RAB with a dedicated RAB-AssignmentRequest that contains
+	 * it may also be that the MSC decides to release the RAB with a dedicated RAB-AssignmentRequest that contains
 	 * a ReleaseList. In this case an FSM will already be present. */
 	if (map->mgw_fi) {
-		/* A RAB Release might be in progress, handle it */
-		rc = handle_rab_release(map, oph, message);
-		if (rc >= 0)
-			return rc;
+		/* Check if the RAB-AssignmentRequest contains a RAB-ReleaseItem that matches the RAB-ID we are
+		 * managing in this HNBGW context map. */
+		if (is_our_rab_release(map, message))
+			return release_mgw_fsm(map, ranap_msg);
 
-		LOGPFSML(map->mgw_fi, LOGL_ERROR,
-			 "mgw_fsm_alloc_and_handle_rab_ass_req() unable to handle RAB-AssignmentRequest!\n");
+		/* The RAB-ReleaseItem in the incoming message should match the RAB ID we are managing. A mismatch may
+		 * mean that there is an inconsistency between the HNBGW and the MSC state and the MGW FSM on the HNBGW
+		 * side may serve an abandonned connection, which we will now close. However we must also assume that
+		 * the incoming message may still contain a RAB-Assignment for a new RTP stream, so we still must
+		 * continue with the message evaluation. */
 		osmo_fsm_inst_state_chg(map->mgw_fi, MGW_ST_FAILURE, 0, 0);
 		OSMO_ASSERT(map->mgw_fi == NULL);
 	}
@@ -718,14 +744,16 @@ int handle_rab_ass_req(struct hnbgw_context_map *map, struct osmo_prim_hdr *oph,
 	 * will block all incoming RAB assignments that try to assign more (or less) than one RAB. */
 	if (ranap_rab_ass_req_ies_get_count(&message->msg.raB_AssignmentRequestIEs) != 1) {
 		LOGP(DMGW, LOGL_ERROR,
-		     "mgw_fsm_alloc_and_handle_rab_ass_req() rua_ctx_id=%d, RAB-AssignmentRequest with more than one RAB assignment -- abort!\n",
-		     map->rua_ctx_id);
+		     "%s() rua_ctx_id=%d, RAB-AssignmentRequest with more than one RAB assignment -- abort!\n",
+		     __func__, map->rua_ctx_id);
 		tx_release_req(map);
 		return -1;
 	}
 
 	mgw_fsm_priv = talloc_zero(map, struct mgw_fsm_priv);
 	mgw_fsm_priv->map = map;
+
+	talloc_steal(mgw_fsm_priv, message);
 	mgw_fsm_priv->ranap_rab_ass_req_message = message;
 
 	/* Allocate FSM */
@@ -738,34 +766,25 @@ int handle_rab_ass_req(struct hnbgw_context_map *map, struct osmo_prim_hdr *oph,
 }
 
 /*! Handlie RANAP RAB AssignmentResponse (deliver message, complete RTP stream switching).
- *  \ptmap[in] map hanbgw context map that is responsible for this call.
- *  \ptmap[in] oph osmo prim header with RANAP RAB AssignmentResponse (function takes ownership).
- *  \ptmap[in] message ranap message container with decoded ranap message (function takes ownership).
+ *  \param[in] map hnbgw context map that is responsible for this call.
+ *  \param[in] ranap_msg msgb containing RANAP RAB AssignmentResponse at msgb_l2(), allocated in OTC_SELECT.
+ *                       This function may talloc_steal(ranap_msg) to keep it for later.
+ *  \param[in] message decoded RANAP message container, allocated in OTC_SELECT.
+ *                     This function may talloc_steal(message) to keep it for later.
  *  \returns 0 on success; negative on error. */
-int mgw_fsm_handle_rab_ass_resp(struct hnbgw_context_map *map, struct osmo_prim_hdr *oph, ranap_message *message)
+int mgw_fsm_handle_rab_ass_resp(struct hnbgw_context_map *map, struct msgb *ranap_msg, ranap_message *message)
 {
 	struct mgw_fsm_priv *mgw_fsm_priv;
-	struct osmo_scu_prim *prim;
-	struct msgb *msg;
 
-	OSMO_ASSERT(oph);
+	OSMO_ASSERT(ranap_msg);
 
 	if (!map->mgw_fi) {
 		/* NOTE: This situation is a corner-case. We may end up here when the co-located MGW caused a problem
 		 * on the way between RANAP RAB Assignment Request and RANAP RAB Assignment Response. */
 
 		LOGP(DMGW, LOGL_ERROR,
-		     "mgw_fsm_handle_rab_ass_resp() rua_ctx_id=%d, no MGW fsm -- sending Iu-Release-Request!\n",
-		     map->rua_ctx_id);
-
-		/* Cleanup ranap message */
-		ranap_cn_rx_co_free(message);
-		talloc_free(message);
-
-		/* Toss RAB-AssignmentResponse */
-		prim = (struct osmo_scu_prim *)oph;
-		msg = prim->oph.msg;
-		msgb_free(msg);
+		     "%s() rua_ctx_id=%d, no MGW fsm -- sending Iu-Release-Request!\n",
+		     __func__, map->rua_ctx_id);
 
 		/* Send a release request, to make sure that the MSC is aware of the problem. */
 		tx_release_req(map);
@@ -773,14 +792,19 @@ int mgw_fsm_handle_rab_ass_resp(struct hnbgw_context_map *map, struct osmo_prim_
 	}
 
 	mgw_fsm_priv = map->mgw_fi->priv;
-	mgw_fsm_priv->ranap_rab_ass_resp_oph = oph;
+
+	talloc_steal(mgw_fsm_priv, ranap_msg);
+	mgw_fsm_priv->ranap_rab_ass_resp_msgb = ranap_msg;
+
+	talloc_steal(mgw_fsm_priv, message);
 	mgw_fsm_priv->ranap_rab_ass_resp_message = message;
+
 	osmo_fsm_inst_dispatch(map->mgw_fi, MGW_EV_RAB_ASS_RESP, NULL);
 	return 0;
 }
 
 /*! Release the FSM and clear its associated RTP streams.
- *  \ptmap[in] map hanbgw context map that is responsible for this call.
+ *  \ptmap[in] map hnbgw context map that is responsible for this call.
  *  \returns 0 on success; negative on error. */
 int mgw_fsm_release(struct hnbgw_context_map *map)
 {
