@@ -45,6 +45,7 @@
 #include <osmocom/hnbgw/hnbgw.h>
 #include <osmocom/hnbgw/context_map.h>
 #include <osmocom/hnbgw/ranap_rab_ass.h>
+#include <osmocom/hnbgw/ps_rab_ass_fsm.h>
 #include <osmocom/hnbgw/ps_rab_fsm.h>
 
 #include <osmocom/hnbgw/hnbgw_rua.h>
@@ -56,13 +57,6 @@
 #define LOG_PS_RAB_ASS(RAB_ASS, LOGL, FMT, ARGS...) \
 	LOGPFSML((RAB_ASS) ? (RAB_ASS)->fi : NULL, LOGL, FMT, ##ARGS)
 
-enum ps_rab_ass_fsm_event {
-	PS_RAB_ASS_EV_LOCAL_F_TEIDS_RX,
-	PS_RAB_ASS_EV_RAB_ASS_RESP,
-	PS_RAB_ASS_EV_RAB_ESTABLISHED,
-	PS_RAB_ASS_EV_RAB_FAIL,
-};
-
 static const struct value_string ps_rab_ass_fsm_event_names[] = {
 	OSMO_VALUE_STRING(PS_RAB_ASS_EV_LOCAL_F_TEIDS_RX),
 	OSMO_VALUE_STRING(PS_RAB_ASS_EV_RAB_ASS_RESP),
@@ -110,7 +104,7 @@ struct ps_rab_ass {
 	ranap_message *ranap_rab_ass_req_message;
 
 	ranap_message *ranap_rab_ass_resp_message;
-	struct osmo_prim_hdr *ranap_rab_ass_resp_oph;
+	struct msgb *ranap_rab_ass_resp_msgb;
 
 	/* A RAB Assignment may contain more than one RAB. Each RAB sets up a distinct ps_rab_fsm (aka PFCP session) and
 	 * reports back about local F-TEIDs assigned by the UPF. This gives the nr of RAB events we expect from
@@ -125,7 +119,7 @@ struct osmo_tdef_state_timeout ps_rab_ass_fsm_timeouts[32] = {
 };
 
 #define ps_rab_ass_fsm_state_chg(state) \
-	osmo_tdef_fsm_inst_state_chg(fi, state, ps_rab_ass_fsm_timeouts, ps_T_defs, -1)
+	osmo_tdef_fsm_inst_state_chg(fi, state, ps_rab_ass_fsm_timeouts, hnbgw_T_defs, -1)
 
 static struct osmo_fsm ps_rab_ass_fsm;
 
@@ -214,20 +208,28 @@ error_exit:
 	return rc;
 }
 
-int hnbgw_gtpmap_rx_rab_ass_req(struct hnbgw_context_map *map, struct osmo_prim_hdr *oph, ranap_message *message)
+/*! Allocate ps_rab_ass_fsm and handle PS RANAP RAB AssignmentRequest.
+ *  \param[in] map hnbgw context map that is responsible for this conn.
+ *  \param[in] ranap_msg msgb containing RANAP RAB AssignmentRequest at msgb_l2(), allocated in OTC_SELECT.
+ *                       This function may talloc_steal(ranap_msg) to keep it for later.
+ *  \param[in] message decoded RANAP message container, allocated in OTC_SELECT.
+ *                     This function may talloc_steal(message) to keep it for later.
+ *  \returns 0 on success; negative on error. */
+int hnbgw_gtpmap_rx_rab_ass_req(struct hnbgw_context_map *map, struct msgb *ranap_msg, ranap_message *message)
 {
 	RANAP_RAB_AssignmentRequestIEs_t *ies = &message->msg.raB_AssignmentRequestIEs;
 	int i;
 
-	struct hnb_gw *hnb_gw = map->hnb_ctx->gw;
 	struct ps_rab_ass *rab_ass;
 	struct osmo_fsm_inst *fi;
 
 	rab_ass = ps_rab_ass_alloc(map);
+
+	talloc_steal(rab_ass, message);
 	rab_ass->ranap_rab_ass_req_message = message;
 	/* Now rab_ass owns message and will clean it up */
 
-	if (!osmo_pfcp_cp_peer_is_associated(hnb_gw->pfcp.cp_peer)) {
+	if (!osmo_pfcp_cp_peer_is_associated(g_hnbgw->pfcp.cp_peer)) {
 		LOG_MAP(map, DLPFCP, LOGL_ERROR, "PFCP is not associated, cannot set up GTP mapping\n");
 		goto no_rab;
 	}
@@ -351,8 +353,10 @@ continue_cleanloop:
 		ps_rab_ass_failure(rab_ass);
 		return;
 	}
-	rua_tx_dt(rab_ass->map->hnb_ctx, rab_ass->map->is_ps, rab_ass->map->rua_ctx_id, msg->data, msg->len);
-	msgb_free(msg);
+	talloc_steal(OTC_SELECT, msg);
+	msg->l2h = msg->data;
+	map_rua_dispatch(rab->map, MAP_RUA_EV_TX_DIRECT_TRANSFER, msg);
+
 	/* The request message has been forwarded. The response will be handled by a new FSM instance.
 	 * We are done. */
 	osmo_fsm_inst_term(rab_ass->fi, OSMO_FSM_TERM_REGULAR, NULL);
@@ -385,7 +389,7 @@ static int ps_rab_setup_access_remote(struct ps_rab_ass *rab_ass,
 	return ps_rab_rx_access_remote_f_teid(map, rab_id, &args);
 }
 
-int hnbgw_gtpmap_rx_rab_ass_resp(struct hnbgw_context_map *map, struct osmo_prim_hdr *oph, ranap_message *message)
+int hnbgw_gtpmap_rx_rab_ass_resp(struct hnbgw_context_map *map, struct msgb *ranap_msg, ranap_message *message)
 {
 	/* hNodeB responds with its own F-TEIDs. Need to tell the UPF about those to complete the GTP mapping.
 	 * 1. here, extract the F-TEIDs (one per RAB),
@@ -408,7 +412,6 @@ int hnbgw_gtpmap_rx_rab_ass_resp(struct hnbgw_context_map *map, struct osmo_prim
 	struct ps_rab_ass *rab_ass;
 	struct osmo_fsm_inst *fi;
 	RANAP_RAB_AssignmentResponseIEs_t *ies;
-	struct hnb_gw *hnb_gw = map->hnb_ctx->gw;
 
 	/* Make sure we indeed deal with a setup-or-modify list */
 	ies = &message->msg.raB_AssignmentResponseIEs;
@@ -418,11 +421,15 @@ int hnbgw_gtpmap_rx_rab_ass_resp(struct hnbgw_context_map *map, struct osmo_prim
 	}
 
 	rab_ass = ps_rab_ass_alloc(map);
+
+	talloc_steal(rab_ass, message);
 	rab_ass->ranap_rab_ass_resp_message = message;
-	rab_ass->ranap_rab_ass_resp_oph = oph;
+
+	talloc_steal(rab_ass, ranap_msg);
+	rab_ass->ranap_rab_ass_resp_msgb = ranap_msg;
 	/* Now rab_ass owns message and will clean it up */
 
-	if (!osmo_pfcp_cp_peer_is_associated(hnb_gw->pfcp.cp_peer)) {
+	if (!osmo_pfcp_cp_peer_is_associated(g_hnbgw->pfcp.cp_peer)) {
 		LOG_PS_RAB_ASS(rab_ass, LOGL_ERROR, "PFCP is not associated, cannot set up GTP mapping\n");
 		ps_rab_ass_failure(rab_ass);
 		return -1;
@@ -494,8 +501,8 @@ static void ps_rab_ass_resp_send_if_ready(struct ps_rab_ass *rab_ass)
 {
 	int i;
 	int rc;
-	struct hnbgw_cnlink *cn = rab_ass->map->cn_link;
 	RANAP_RAB_AssignmentResponseIEs_t *ies = &rab_ass->ranap_rab_ass_resp_message->msg.raB_AssignmentResponseIEs;
+	struct msgb *msg;
 
 	/* Go through all RABs in the RAB Assignment Response message and replace with the F-TEID that the UPF assigned,
 	 * verifying that instructing the UPF has succeeded. */
@@ -532,7 +539,15 @@ static void ps_rab_ass_resp_send_if_ready(struct ps_rab_ass *rab_ass)
 			return;
 		}
 
-		/* Replace GTP endpoint */
+		/* Replace GTP endpoint.
+		 * memory: ranap_new_transp_layer_addr() frees previous buffer in transportLayerAddress, if any. The
+		 * entire rab_item is freed along with item_ies.
+		 *
+		 * BTW, If I free this explicitly below with
+		 *   ASN_STRUCT_FREE_CONTENTS_ONLY(asn_DEF_BIT_STRING, &rab_item->transportLayerAddress);
+		 * that causes a leak in talloc_asn1_ctx: "iu_helpers.c:205 contains  20 bytes in  1 blocks".
+		 * I couldn't figure out why, but things are freed properly when leaving it all up to item_ies.
+		 */
 		if (ranap_new_transp_layer_addr(rab_item->transportLayerAddress, &rab->core.local.addr,
 						rab->access.use_x213_nsap) < 0) {
 			ASN_STRUCT_FREE_CONTENTS_ONLY(asn_DEF_RANAP_RAB_SetupOrModifiedItem, &item_ies);
@@ -549,6 +564,8 @@ static void ps_rab_ass_resp_send_if_ready(struct ps_rab_ass *rab_ass)
 
 		teid_be = htonl(rab->core.local.teid);
 		rab_item->iuTransportAssociation->present = RANAP_IuTransportAssociation_PR_gTP_TEI;
+		/* memory: OCTET_STRING_fromBuf() frees previous buffer in gTP_TEI, if any. The entire rab_item is freed
+		 * along with item_ies. */
 		OCTET_STRING_fromBuf(&rab_item->iuTransportAssociation->choice.gTP_TEI,
 				     (const char *)&teid_be, sizeof(teid_be));
 
@@ -565,10 +582,13 @@ continue_cleanloop:
 		ASN_STRUCT_FREE_CONTENTS_ONLY(asn_DEF_RANAP_RAB_SetupOrModifiedItem, &item_ies);
 	}
 
+	msg = rab_ass->ranap_rab_ass_resp_msgb;
+	rab_ass->ranap_rab_ass_resp_msgb = NULL;
+	talloc_steal(OTC_SELECT, msg);
+
 	/* Replaced all the GTP info, re-encode the message. Since we are replacing data 1:1, taking care to use the
 	 * same IP address encoding, the resulting message size must be identical to the original message size. */
-	rc = ranap_rab_ass_resp_encode(msgb_l2(rab_ass->ranap_rab_ass_resp_oph->msg),
-				       msgb_l2len(rab_ass->ranap_rab_ass_resp_oph->msg), ies);
+	rc = ranap_rab_ass_resp_encode(msgb_l2(msg), msgb_l2len(msg), ies);
 	if (rc < 0) {
 		LOG_PS_RAB_ASS(rab_ass, LOGL_ERROR, "Re-encoding RANAP PS RAB-AssignmentResponse failed\n");
 		ps_rab_ass_failure(rab_ass);
@@ -576,8 +596,8 @@ continue_cleanloop:
 	}
 
 	LOG_PS_RAB_ASS(rab_ass, LOGL_NOTICE, "Sending RANAP PS RAB-AssignmentResponse with mapped GTP info\n");
-	rc = osmo_sccp_user_sap_down(cn->sccp_user, rab_ass->ranap_rab_ass_resp_oph);
-	rab_ass->ranap_rab_ass_resp_oph = NULL;
+
+	rc = map_sccp_dispatch(rab_ass->map, MAP_SCCP_EV_TX_DATA_REQUEST, msg);
 	if (rc < 0) {
 		LOG_PS_RAB_ASS(rab_ass, LOGL_ERROR, "Sending RANAP PS RAB-AssignmentResponse failed\n");
 		ps_rab_ass_failure(rab_ass);
@@ -590,29 +610,8 @@ continue_cleanloop:
 static void ps_rab_ass_fsm_cleanup(struct osmo_fsm_inst *fi, enum osmo_fsm_term_cause cause)
 {
 	struct ps_rab_ass *rab_ass = fi->priv;
-	struct osmo_scu_prim *scu_prim;
-	struct msgb *scu_msg;
 	struct ps_rab *rab;
 
-	if (rab_ass->ranap_rab_ass_req_message) {
-		ranap_ran_rx_co_free(rab_ass->ranap_rab_ass_req_message);
-		talloc_free(rab_ass->ranap_rab_ass_req_message);
-		rab_ass->ranap_rab_ass_req_message = NULL;
-	}
-
-	if (rab_ass->ranap_rab_ass_resp_message) {
-		ranap_cn_rx_co_free(rab_ass->ranap_rab_ass_resp_message);
-		talloc_free(rab_ass->ranap_rab_ass_resp_message);
-		rab_ass->ranap_rab_ass_resp_message = NULL;
-	}
-
-	if (rab_ass->ranap_rab_ass_resp_oph) {
-		scu_prim = (struct osmo_scu_prim *)rab_ass->ranap_rab_ass_resp_oph;
-		scu_msg = scu_prim->oph.msg;
-		msgb_free(scu_msg);
-		rab_ass->ranap_rab_ass_resp_oph = NULL;
-	}
-
 	llist_for_each_entry(rab, &rab_ass->map->ps_rabs, entry) {
 		if (rab->req_fi == fi)
 			rab->req_fi = NULL;
@@ -620,6 +619,7 @@ static void ps_rab_ass_fsm_cleanup(struct osmo_fsm_inst *fi, enum osmo_fsm_term_
 			rab->resp_fi = NULL;
 	}
 
+	/* remove from map->ps_rab_ass */
 	llist_del(&rab_ass->entry);
 }
 
