@@ -41,11 +41,9 @@ struct msgb *ranap_rab_ass_req_encode(RANAP_RAB_AssignmentRequestIEs_t *rab_assi
 {
 	int rc;
 	struct msgb *msg;
-	RANAP_RAB_AssignmentRequest_t _rab_assignment_request;
+	RANAP_RAB_AssignmentRequest_t _rab_assignment_request = { 0 };
 	RANAP_RAB_AssignmentRequest_t *rab_assignment_request = &_rab_assignment_request;
 
-	memset(rab_assignment_request, 0, sizeof(*rab_assignment_request));
-
 	rc = ranap_encode_rab_assignmentrequesties(rab_assignment_request, rab_assignment_request_ies);
 	if (rc < 0)
 		return NULL;
@@ -72,11 +70,10 @@ int ranap_rab_ass_resp_encode(uint8_t *data, unsigned int len,
 	int rc;
 	struct msgb *msg;
 
-	RANAP_RAB_AssignmentResponse_t _rab_assignment_response;
+	RANAP_RAB_AssignmentResponse_t _rab_assignment_response = { 0 };
 	RANAP_RAB_AssignmentResponse_t *rab_assignment_response = &_rab_assignment_response;
 
 	memset(data, 0, len);
-	memset(rab_assignment_response, 0, sizeof(*rab_assignment_response));
 
 	rc = ranap_encode_rab_assignmentresponseies(rab_assignment_response, rab_assignment_response_ies);
 	if (rc < 0)
@@ -87,7 +84,7 @@ int ranap_rab_ass_resp_encode(uint8_t *data, unsigned int len,
 				     RANAP_Criticality_reject,
 				     &asn_DEF_RANAP_RAB_AssignmentResponse, rab_assignment_response);
 
-	/* 'msg' has been generated, we cann now release the input 'out' */
+	/* 'msg' has been generated, we can now release the input 'rab_assignment_response' */
 	ASN_STRUCT_FREE_CONTENTS_ONLY(asn_DEF_RANAP_RAB_AssignmentResponse, rab_assignment_response);
 
 	if (!msg)
@@ -176,7 +173,9 @@ static RANAP_IE_t *failed_list_item_from_rab_ass_resp(const RANAP_RAB_Assignment
 	return ies->raB_FailedList.raB_FailedList_ies.list.array[index];
 }
 
-/* find the RAB specified by rab_id in ies and when found, decode the result into items_ies */
+/* Find the RAB specified by rab_id in ies, decode the result into items_ies and return a positive index.
+ * The caller is responsible for freeing the contents in items_ies. In case of failure, the return code
+ * will be negative. */
 static int decode_rab_smditms_from_resp_ies(RANAP_RAB_SetupOrModifiedItemIEs_t *items_ies,
 					    RANAP_RAB_AssignmentResponseIEs_t *ies, uint8_t rab_id)
 {
@@ -313,7 +312,7 @@ int ranap_rab_ass_req_ies_extract_inet_addr(struct osmo_sockaddr *addr, uint8_t
 					    RANAP_RAB_AssignmentRequestIEs_t *ies, unsigned int index)
 {
 	RANAP_ProtocolIE_FieldPair_t *protocol_ie_field_pair;
-	RANAP_RAB_SetupOrModifyItemFirst_t _rab_setup_or_modify_item_first;
+	RANAP_RAB_SetupOrModifyItemFirst_t _rab_setup_or_modify_item_first = { 0 };
 	RANAP_RAB_SetupOrModifyItemFirst_t *rab_setup_or_modify_item_first = &_rab_setup_or_modify_item_first;
 	RANAP_TransportLayerAddress_t *trasp_layer_addr;
 	RANAP_IuTransportAssociation_t *transp_assoc;
@@ -382,7 +381,7 @@ error:
  *  \returns 0 on success; negative on error. */
 int ranap_rab_ass_resp_ies_extract_inet_addr(struct osmo_sockaddr *addr, RANAP_RAB_AssignmentResponseIEs_t *ies, uint8_t rab_id)
 {
-	RANAP_RAB_SetupOrModifiedItemIEs_t _rab_setup_or_modified_items_ies;
+	RANAP_RAB_SetupOrModifiedItemIEs_t _rab_setup_or_modified_items_ies = { 0 };
 	RANAP_RAB_SetupOrModifiedItemIEs_t *rab_setup_or_modified_items_ies = &_rab_setup_or_modified_items_ies;
 	RANAP_RAB_SetupOrModifiedItem_t *rab_setup_or_modified_item;
 	uint16_t port;
@@ -434,7 +433,7 @@ error:
 int ranap_rab_ass_req_ies_replace_inet_addr(RANAP_RAB_AssignmentRequestIEs_t *ies, struct osmo_sockaddr *addr, uint8_t rab_id)
 {
 	RANAP_ProtocolIE_FieldPair_t *protocol_ie_field_pair;
-	RANAP_RAB_SetupOrModifyItemFirst_t _rab_setup_or_modify_item_first;
+	RANAP_RAB_SetupOrModifyItemFirst_t _rab_setup_or_modify_item_first = { 0 };
 	RANAP_RAB_SetupOrModifyItemFirst_t *rab_setup_or_modify_item_first = &_rab_setup_or_modify_item_first;
 	RANAP_TransportLayerInformation_t *old_transport_layer_information = NULL;
 	RANAP_TransportLayerInformation_t *new_transport_layer_information = NULL;
@@ -504,7 +503,7 @@ error:
 int ranap_rab_ass_resp_ies_replace_inet_addr(RANAP_RAB_AssignmentResponseIEs_t *ies, struct osmo_sockaddr *addr, uint8_t rab_id)
 {
 	RANAP_IE_t *setup_or_modified_list_ie;
-	RANAP_RAB_SetupOrModifiedItemIEs_t _rab_setup_or_modified_items_ies;
+	RANAP_RAB_SetupOrModifiedItemIEs_t _rab_setup_or_modified_items_ies = { 0 };
 	RANAP_RAB_SetupOrModifiedItemIEs_t *rab_setup_or_modified_items_ies = &_rab_setup_or_modified_items_ies;
 	RANAP_RAB_SetupOrModifiedItem_t *rab_setup_or_modified_item;
 	RANAP_TransportLayerInformation_t *temp_transport_layer_information = NULL;
@@ -573,20 +572,19 @@ error:
  *  \returns true when RAB could be identified as failed; false otherwise */
 bool ranap_rab_ass_resp_ies_check_failure(RANAP_RAB_AssignmentResponseIEs_t *ies, uint8_t rab_id)
 {
-	RANAP_RAB_FailedItemIEs_t _rab_failed_items_ies;
+	RANAP_RAB_FailedItemIEs_t _rab_failed_items_ies = { 0 };
 	RANAP_RAB_FailedItemIEs_t *rab_failed_items_ies = &_rab_failed_items_ies;
 	int rc;
-	bool result = true;
 
-	/* If we can get a failed item for the specified RAB ID, then we know that the
+	/* If we can get a failed item (rc >= 0) for the specified RAB ID, then we know that the
 	 * HNB reported the RAB Assignment as failed */
 	rc = decode_rab_flitms_from_resp_ies(rab_failed_items_ies, ies, rab_id);
 	if (rc < 0)
-		result = false;
+		return false;
 
 	ASN_STRUCT_FREE_CONTENTS_ONLY(asn_DEF_RANAP_RAB_FailedItem, rab_failed_items_ies);
 
-	return result;
+	return true;
 }
 
 /*! Check if a specific RAB is present in an RAB-ReleaseList inside RANAP_RAB_AssignmentRequestIEs.
@@ -595,7 +593,7 @@ bool ranap_rab_ass_resp_ies_check_failure(RANAP_RAB_AssignmentResponseIEs_t *ies
  *  \returns true when RAB is intended for release; false otherwise */
 bool ranap_rab_ass_req_ies_check_release(RANAP_RAB_AssignmentRequestIEs_t *ies, uint8_t rab_id)
 {
-	RANAP_RAB_ReleaseItemIEs_t _rab_release_items_ies;
+	RANAP_RAB_ReleaseItemIEs_t _rab_release_items_ies = { 0 };
 	RANAP_RAB_ReleaseItemIEs_t *rab_release_items_ies = &_rab_release_items_ies;
 	int rc;
 	bool result = true;
