@@ -33,7 +33,6 @@
 #include <osmocom/ranap/ranap_common_cn.h>
 #include <osmocom/ranap/ranap_common_ran.h>
 
-static void *tall_hnb_ctx;
 static void *msgb_ctx;
 extern void *talloc_asn1_ctx;
 
@@ -86,7 +85,7 @@ void test_ranap_rab_ass_resp_decode_encode(void)
 	};
 	uint8_t encoded[sizeof(testvec)];
 
-	rc = ranap_cn_rx_co_decode(talloc_asn1_ctx, &message, testvec, sizeof(testvec));
+	rc = ranap_cn_rx_co_decode2(&message, testvec, sizeof(testvec));
 	OSMO_ASSERT(rc == 0);
 
 	rc = ranap_rab_ass_resp_encode(encoded, sizeof(encoded), &message.msg.raB_AssignmentResponseIEs);
@@ -145,7 +144,7 @@ void test_ranap_rab_ass_resp_ies_extract_inet_addr(void)
 		0x00, 0x40, 0x04, 0x0a, 0x00, 0x00
 	};
 
-	rc = ranap_cn_rx_co_decode(talloc_asn1_ctx, &message, testvec, sizeof(testvec));
+	rc = ranap_cn_rx_co_decode2(&message, testvec, sizeof(testvec));
 	OSMO_ASSERT(rc == 0);
 	rc = ranap_rab_ass_resp_ies_extract_inet_addr(&addr, &message.msg.raB_AssignmentResponseIEs, 7);
 	osmo_sockaddr_str_from_sockaddr(&addr_str, &addr.u.sas);
@@ -245,7 +244,7 @@ void test_ranap_rab_ass_resp_ies_replace_inet_addr(void)
 		0x00, 0x40, 0x04, 0xd2, 0x00, 0x00
 	};
 
-	rc = ranap_cn_rx_co_decode(talloc_asn1_ctx, &message, testvec_in, sizeof(testvec_in));
+	rc = ranap_cn_rx_co_decode2(&message, testvec_in, sizeof(testvec_in));
 	OSMO_ASSERT(rc == 0);
 
 	rc = ranap_rab_ass_resp_ies_extract_inet_addr(&addr, &message.msg.raB_AssignmentResponseIEs, 6);
@@ -285,7 +284,7 @@ void test_ranap_rab_ass_resp_ies_check_failure(void)
 		0x40, 0x03, 0x05, 0xd0, 0x00
 	};
 
-	rc = ranap_cn_rx_co_decode(talloc_asn1_ctx, &message, testvec, sizeof(testvec));
+	rc = ranap_cn_rx_co_decode2(&message, testvec, sizeof(testvec));
 	OSMO_ASSERT(rc == 0);
 
 	rab_failed_at_hnb =
@@ -300,6 +299,31 @@ void test_ranap_rab_ass_resp_ies_check_failure(void)
 	ranap_cn_rx_co_free(&message);
 }
 
+/* Same as above but with a message that does not even contain a FailedItemsIEs list */
+void test_ranap_rab_ass_resp_ies_check_failure_no_faileditemies(void)
+{
+	int rc;
+	ranap_message message;
+	bool rab_failed_at_hnb;
+	uint8_t testvec[] = {
+		0x60, 0x00, 0x00, 0x2a, 0x00, 0x00, 0x01, 0x00,
+		0x34, 0x40, 0x23, 0x00, 0x00, 0x01, 0x00, 0x33,
+		0x40, 0x1c, 0x60, 0x3a, 0x7c, 0x35, 0x00, 0x01,
+		0x0a, 0x09, 0x01, 0xa4, 0x00, 0x00, 0x00, 0x00,
+		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
+		0x00, 0x40, 0x04, 0x0a, 0x00, 0x00
+	};
+
+	rc = ranap_cn_rx_co_decode2(&message, testvec, sizeof(testvec));
+
+	OSMO_ASSERT(rc == 0);
+	rab_failed_at_hnb =
+		ranap_rab_ass_resp_ies_check_failure(&message.msg.raB_AssignmentResponseIEs, 23);
+	OSMO_ASSERT(rab_failed_at_hnb == 0)
+
+	ranap_cn_rx_co_free(&message);
+}
+
 void test_ranap_rab_ass_req_ies_check_release(void)
 {
 	int rc;
@@ -368,12 +392,13 @@ static const struct log_info test_log_info = {
 int test_init(void)
 {
 	int rc;
+	void *root_ctx;
 
-	tall_hnb_ctx = talloc_named_const(NULL, 0, "hnb_context");
-	msgb_ctx = msgb_talloc_ctx_init(NULL, 0);
-	talloc_asn1_ctx = talloc_named_const(NULL, 0, "asn1_context");
+	root_ctx = talloc_named_const(NULL, 0, "ranap_rab_ass_test");
+	msgb_ctx = msgb_talloc_ctx_init(root_ctx, 0);
+	talloc_asn1_ctx = talloc_named_const(root_ctx, 0, "asn1_context");
 
-	rc = osmo_init_logging2(tall_hnb_ctx, &test_log_info);
+	rc = osmo_init_logging2(root_ctx, &test_log_info);
 	if (rc < 0)
 		exit(1);
 
@@ -413,6 +438,7 @@ int main(int argc, char **argv)
 	test_ranap_rab_ass_req_ies_replace_inet_addr();
 	test_ranap_rab_ass_resp_ies_replace_inet_addr();
 	test_ranap_rab_ass_resp_ies_check_failure();
+	test_ranap_rab_ass_resp_ies_check_failure_no_faileditemies();
 	test_ranap_rab_ass_req_ies_check_release();
 	test_ranap_rab_ass_req_ies_get_count();
 
